﻿//--------------------------------------------------------------
// Dataset.fs           Dataset which encapsulates the data, allowing online reading.
//
// 2008 written by Xinhua Zhang
// Microsoft Research Ltd
//--------------------------------------------------------------

#light

module MSRC.Dataset



open System
open System.IO
open System.Collections.Generic
open System.Diagnostics
open MSRC.Inference
open System.Text.RegularExpressions

//open MSRC.Inference.Collections
//open MSRC.Inference.Distributions
open MSRC.Inference.Distributions.Gaussians
//open MSRC.Inference.Schedule



type DebugParam (_showWeightEachTrain:bool, _showPostEachTrain:bool, 
                    _printTrainData:bool, _printTestData:bool, _printLabelIdToName:bool,
                    _saveFinalModel:bool, _saveTestPosterior:bool,
                    _debugDumpDir:string) = 

    member this.showWeightEachTrain with get () = _showWeightEachTrain
    member this.showPostEachTrain with get () = _showPostEachTrain
    member this.printTrainData with get () = _printTrainData
    member this.printTestData with get () = _printTestData
    member this.printLabelIdToName with get () = _printLabelIdToName
    member this.saveFinalModel with get () = _saveFinalModel
    member this.saveTestPosterior with get () = _saveTestPosterior
    member this.debugDumpDir with get () = _debugDumpDir

    member this.print () = 
        printfn "showWeightEachTrain = %A\tshowPostEachTrain = %A" _showWeightEachTrain _showPostEachTrain
        printfn "printTrainData = %A\tprintTestData = %A" _printTrainData _printTestData
        printfn "printLabelIdToName = %A" _printLabelIdToName
        printfn "saveFinalModel = %A\tsaveTestPosterior = %A" _saveFinalModel _saveTestPosterior
        printfn "Debug dump directory = %s" _debugDumpDir
        printfn "--------------------------------------------------------------------"


/// A bag of model parameters
type ModelParameter<[<Measure>] 'a> (   _datasetName:string,                                         
                                        _filterFname:string,
                                        _root:string,
                                        
                                        _monitorInterval:int,
                                        
                                        _betaSq:float,
                                        _epsilon:float,
                                        _delta:float,

                                        _useBias:bool,
                                        _biasValue:float,
                                        _normalizeFeatures:bool,
                                        _featureCutoffValue:float,

                                        _numBill:int, 
                                        _numTrSweep:int,
                                        
                                        _useValid:bool,
                                        _useCMN:bool,
                                        _showEnum:bool, 
                                        _corDiag:bool, 

                                        _maxRandValue:int,
                                        _noSample:int, 
                                        _noThreshGridStretch:int,
                                        _sampleBarrier:float,
                                        _maxPostSpace:int,
                                        _useFastThreshSet:bool,
                                        
                                        _writeSampleExpfs:bool, 
                                        _writeRefFscore:bool,  

                                        _weightPriorSigma:float, 
                                        _biasPriorSigma:float,                                        
                                        _threshPriorSigma:float,
                                        _useGlobalBias:bool, 
                                        
                                        _labelNoiseRho:float,
                                        _evidenceIter:int,
                                        
                                        _useEP:bool,
                                        _useClosed:bool, 
                                        
                                        _threshScaling:float
                                        
                                      ) =
 
    let innerMonitorInterval = ref _monitorInterval

    member this.datasetName with get () = _datasetName

    member this.filterFname with get () = _filterFname
    
    member this.root with get () = _root

    member this.monitorInterval with get () = !innerMonitorInterval
                                 
    member this.updateMonitorInterval value = 
                                            if _monitorInterval = -1 then 
                                                innerMonitorInterval := value * _numTrSweep 
    
    member this.betaSquare with get () : float<'a^2> = _betaSq * 1.0<_>
    member this.epsilon with get () : float<'a> = _epsilon * 1.0<_>
    member this.delta with get () : float</'a> = _delta * 1.0<_>

    member this.useBias with get () = _useBias
    member this.biasValue with get () = _biasValue
    member this.normalizeFeatures with get () = _normalizeFeatures
    member this.featureCutoffValue with get () = _featureCutoffValue

    member this.numBill with get () = _numBill
    member this.numTrSweep with get () = _numTrSweep

    member this.useValid with get () = _useValid
    member this.useCMN with get () = _useCMN
    member this.showEnum with get () = _showEnum
    member this.corDiag with get () = _corDiag

    member this.maxRandValue with get () = _maxRandValue
    member this.noSample with get () = _noSample
    member this.noThreshGridStretch with get () = _noThreshGridStretch
    member this.sampleBarrier with get () = _sampleBarrier  
    member this.maxPostSpace with get () = _maxPostSpace
    member this.useFastThreshSet with get () = _useFastThreshSet    
          
    member this.writeSampleExpFs with get () = _writeSampleExpfs
    member this.writeRefFscore with get () = _writeRefFscore    

    member this.weightPriorMu with get () : float<'a> = 0.0<_>
    member this.weightPriorSigma with get () : float<'a> = _weightPriorSigma * 1.0<_>
    
    member this.biasPriorMu with get () : float<'a> = 0.0<_>
    member this.biasPriorSigma with get () : float<'a> = _biasPriorSigma * 1.0<_>
    
    member this.useGlobalBias with get () = _useGlobalBias
    member this.thresholdPriorMu with get () : float<'a> = 0.0<_>
    member this.thresholdPriorSigma with get () : float<'a> = _threshPriorSigma * 1.0<_>
           
    member this.labelNoiseRho with get () = _labelNoiseRho
    member this.evidenceIter with get () = _evidenceIter
                  
    member this.useEP with get () = _useEP
    member this.useClosed with get () = _useClosed
    
    member this.threshScaling with get () = _threshScaling
                  
    member this.getIDString () = 
        sprintf "%s_%s_bs=%g_wsig=%g_bsig=%g_gsig=%g_nsam=%d" _datasetName _filterFname 
                    (float _betaSq) (float _weightPriorSigma) _biasPriorSigma _threshPriorSigma _noSample
                
    member this.print() = 

        printfn "Dumping parameter settings."

        printfn "Dataset = %s" _datasetName
        
        printfn "filter filename = %s" _filterFname
        
        printfn "betaSquare = %g, epsilon = %g, delta = %g, monitor = %d" _betaSq _epsilon _delta this.monitorInterval

        printfn "use local bias = %A, bias value = %g, normlize feature = %A" _useBias _biasValue _normalizeFeatures
        
        printfn "numBill = %d, numTrainSweep = %d" _numBill _numTrSweep

        printfn "weightPriorSigma = %g, biasPriorSigma = %g, thresholdPriorSigma = %g" _weightPriorSigma _biasPriorSigma _threshPriorSigma
        printfn "use global threshold = %A" _useGlobalBias
                            
        printfn "#Sample = %d, maxRandValue = %d, WriteSampleExpFs = %A, WriteRefFscore = %A"
                    _noSample _maxRandValue _writeSampleExpfs _writeRefFscore
        
        printfn "labelNoiseRho = %g" _labelNoiseRho
        
        printfn "useEP = %A\tuseClosed = %A" _useEP _useClosed
                                    
        printfn "--------------------------------------------------------------------"


/// Data structure for the meta data (properties of the dataset)
type MetaData (debugInfo:DebugParam) =    

    let labelNameToId  = new Dictionary<string, int> ()
    let labelIdToName  = new Dictionary<int, string> ()
    let featureNameToId = new Dictionary<string, int> ()
    let featureIdToName = new Dictionary<int, string> ()
    let labelCount = new ResizeArray<int> ()
    let noInstance = ref 0
    let noTestInstance = ref 0

    member this.LabelNameToId = labelNameToId
    member this.LabelIdToName = labelIdToName
    member this.FeatureNameToId = featureNameToId
    member this.FeatureIdToName = featureIdToName
    member this.LabelCount = labelCount

    member this.NoLabel = labelNameToId.Count
    member this.NoFeature = featureNameToId.Count

    member this.print () = 
                printfn "Meta Data based on training data"
                printfn "num label = %d, num feature = %d" this.NoLabel this.NoFeature
                if debugInfo.printLabelIdToName then
                    printfn "Label ID -> Original Name:"            
                    labelIdToName |> Seq.iter (fun kvp -> printfn "%d -> %s" kvp.Key kvp.Value)            



/// Data structure for one single data point
/// Constructed by (line : string, normalize : bool, metaData : MetaData)
type DataPointSparse <[<Measure>] 'a> (yline : string, xline : string, normalize : bool, 
                                        metaData : MetaData, modelParam : ModelParameter<'a>) =     

    (*  activeFeature: array<(int * float)> 
                        an array which stores the features of the data point
                        each feature is a pair (id : int, value : float). id is from 0
        activeLabels: array<int> 
                        an array  which stores the active labels (from 0)
    *)         
    let (activeLabels, labelOrderToId, labelIdToOrder, activeFeatures, noActiveLabel, noActiveFeature) = 
    
        let labelNameToId = metaData.LabelNameToId
        let featureNameToId = metaData.FeatureNameToId
        
        let (labelOrderToId, labelIdToOrder) = 
            match carelabelOrderToId with
            | false -> (null, null)
            | _ -> (new Dictionary<_,_> (), new Dictionary<_,_> ())

        /////////////////////////////////////////
        // First process the valid labels     
        let activeLabels = 
            [|   
              // need to remove all the labels that did not appear in the metaData
              for s in yline.Trim().Split [| ',' |] do                      
                  if not (String.IsNullOrEmpty s) && labelNameToId.ContainsKey s then
                      let label = labelNameToId.[s]
                      if carelabelOrderToId then
                          let order = labelOrderToId.Count
                          labelOrderToId.Add (order, label)
                          labelIdToOrder.Add (label, order)
                      yield label
            |]
        
        // Sort the labels in ascending order, to accelerate comparing labels during testing
        // From this point, activeLabels may be inconsistent with labelOrderToId and labelIdToOrder
        // This is deliberate, in order to retain original order information
        activeLabels |> Array.sort (fun a b -> if a > b then 1 elif a < b then -1 else 0) 
        
        
        // second process the features          
        let activeFeatures_baby = [| for pair in xline.Trim().Split [| ' ' |] |> Array.map (fun s -> s.Split [| ':' |] ) do
                                        if Math.Abs (float pair.[1]) > modelParam.featureCutoffValue 
                                            && featureNameToId.ContainsKey pair.[0] then
                                            yield (featureNameToId.[pair.[0]], float pair.[1])
                                  |]

        let totalStd = activeFeatures_baby |> Array.fold_left (fun acc (id, value) -> acc + value * value) 0.0 |> Math.Sqrt
        
        let activeFeatures = 
            if normalize then 
                activeFeatures_baby |> Array.map (fun (id, value) -> (id, value / totalStd))
            else
                activeFeatures_baby
        
//        printfn "<<<"
//        activeFeatures_baby |> Seq.iter (fun (id, value) -> printf "%d:%A " id value)
//        printfn ">>>"
        // Finally return the labels, features, and their lengths of this data point        
        (activeLabels, labelOrderToId, labelIdToOrder, activeFeatures, activeLabels.Length, activeFeatures.Length)


    member this.Labels = activeLabels
    member this.Features = activeFeatures
    member this.NoLabel = noActiveLabel
    member this.NoFea = noActiveFeature
    member this.LabelIdToOrder = labelIdToOrder
    
    member this.ContainsLabel (label:int) = 
        
        // Use binary search to check whether the current data point contains this label 
        if noActiveLabel = 0 || activeLabels.[0] > label || activeLabels.[noActiveLabel-1] < label then
            false
        else
            let mutable low = 0
            let mutable high = noActiveLabel - 1
            let mutable found = false
            
            while low <= high do
                let mid = (low + high) / 2
                if activeLabels.[mid] > label then
                    high <- mid - 1
                elif activeLabels.[mid] < label then
                    low <- mid + 1
                else
                    low <- high + 1
                    found <- true
            found
            

    /// Print out the content of the current data point in a user friendly format
    member this.print (labelIdToName : Dictionary<_,_>, featureIdToName : Dictionary<_,_>) = 
        // first print lables
        
        printf "There are %d labels and %d features in all.\n" noActiveLabel noActiveFeature
        
        activeLabels |> Array.iter (fun a ->
            printf "%s," labelIdToName.[a] )
        printf "\b "
        
        activeFeatures |> Array.iter (fun (id, value) ->
            printf "%s:%.2g " featureIdToName.[id] value )
        
        printf "\n"
                
    /// Save the labels and features to file (they have already been sorted when reading)
    member this.sortAndSaveToFile(labelwriter : StreamWriter, featurewriter : StreamWriter) =
    
        activeLabels.[0..(activeLabels.Length-2)] |> Array.iter (fun id -> labelwriter.Write (sprintf "%d," id))
        labelwriter.Write (sprintf "%d " activeLabels.[activeLabels.Length-1])
                        
        activeFeatures |> Array.iter (fun (id, value) ->
            featurewriter.Write(sprintf "%d:%.6f " id value))

        labelwriter.Write("\n")
        featurewriter.Write("\n")
        

        
    /// (cosine) Normalize the features
    member this.normalize() =
        let totalVar = activeFeatures |> Array.fold_left (fun acc (id, value) -> acc + value * value) 0.0
        let totalStd = sqrt totalVar
        
        for i in 0 .. (noActiveFeature - 1) do
            let (id, value) = activeFeatures.[i]
            activeFeatures.[i] <- (id, value / totalStd)
             
             
    /// In the initial scan, just update the meta-data
    static member updateMeta (line_y : string, line_x : string, metaData : MetaData, 
                                modelParam : ModelParameter<'a>, numPos:Dictionary<int, int>) =
    
        let labelNameToId = metaData.LabelNameToId
        let labelIdToName = metaData.LabelIdToName
        let featureNameToId = metaData.FeatureNameToId
        let featureIdToName = metaData.FeatureIdToName

        let labelOrderToId = new Dictionary<_,_> ()
        let labelIdToOrder = new Dictionary<_,_> ()

        // First, labels
        line_y.Trim().Split [| ',' |] |> Array.iter (fun s ->
        
            if not (String.IsNullOrEmpty s) then   // && s = "213"
               match labelNameToId.TryGetValue(s) with
               | (false, _) -> 
                   let id = labelNameToId.Count
                   labelNameToId.Add(s, id)
                   labelIdToName.Add(id, s)
                   metaData.LabelCount.Add (1)
                   numPos.Add(id, 1)
               | (true, id) ->
                   metaData.LabelCount.[id] <- metaData.LabelCount.[id] + 1
                   numPos.[id] <- numPos.[id] + 1
        )

        // Second, features                        
        line_x.Trim().Split [| ' ' |] |> Array.map (fun s -> s.Split [| ':' |] )
                                      |> Array.filter (fun pair -> Math.Abs (float pair.[1]) > modelParam.featureCutoffValue)
                                      |> Array.iter (fun pair ->
                                                         if not (featureNameToId.ContainsKey pair.[0]) then 
                                                              let id = featureNameToId.Count
                                                              featureNameToId.Add(pair.[0], id)
                                                              featureIdToName.Add(id, pair.[0]))
                               
        

type SeqDataSet <[<Measure>] 'a> (filename_x:String, filename_y:String, filename_filter:String, 
                                    normalize:bool, updateMeta:bool, selToken:int, 
                                    metaData:MetaData, modelParam : ModelParameter<'a>) = 

    /// Though no longer useful for training, still some useful statistics for whatever use
    /// Go through the data 

    let curInstance = ref 0     // id of the instance that is currently pointed to
    let totalInFile = ref 0     // total number of instances in the source file
    
    let _numPos = new Dictionary<int, int> ()

    
    let print_warning str =
        printfn "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"
        printfn str
        printfn "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"

    let (noInstance, reader_x, reader_y, reader_filter) = 
            
            let noSelected = ref 0
            
            printfn "Using filter %s" filename_filter
            printfn "Reading feature file: %s\nlabel file %s" filename_x filename_y            

            // We allow filter file to have fewer lines than feature/label files
            if updateMeta then
                printf "Parsing data... "
                let read_x = new StreamReader(File.OpenRead(filename_x))
                let read_y = new StreamReader(File.OpenRead(filename_y))
                let read_filter = new StreamReader(File.OpenRead(filename_filter))
                
                let line_skip = ref 0
                while not read_filter.EndOfStream do

                    totalInFile := !totalInFile + 1
                    if selToken = Int32.of_string (read_filter.ReadLine()) then

                        while not read_x.EndOfStream && not read_y.EndOfStream && !line_skip <> 0 do                            
                                line_skip := !line_skip - 1
                                read_x.ReadLine() |> ignore
                                read_y.ReadLine() |> ignore
                        
                        if !line_skip = 0 then
                            if not read_x.EndOfStream && not read_y.EndOfStream then
                                let line_y = read_y.ReadLine()
                                let line_x = 
                                    if modelParam.useBias then
                                        read_x.ReadLine() + " b:" + (Float.to_string modelParam.biasValue)
                                    else
                                        read_x.ReadLine() 
                                DataPointSparse.updateMeta (line_y, line_x, metaData, modelParam, _numPos)
                                noSelected := !noSelected + 1
                            else
                                print_warning("Filter file is longer than feature/label data file");
                    else
                        line_skip := !line_skip + 1
                
                totalInFile := !totalInFile - !line_skip

                while not read_y.EndOfStream do
                
                    read_y.ReadLine() |> ignore
                    totalInFile := !totalInFile + 1
                
                read_x.Close ()
                read_y.Close () 
                read_filter.Close ()
                
                printfn "done."
                printfn "%d records selected from %d" !noSelected !totalInFile

            else
                let read_filter = new StreamReader(File.OpenRead(filename_filter))
                while not read_filter.EndOfStream do
                    totalInFile := !totalInFile + 1
                    if selToken = Int32.of_string (read_filter.ReadLine()) then
                        noSelected := !noSelected + 1

                read_filter.Close ()

                printfn "Parsing skipped. "
                printfn "%d records selected from %d" !noSelected !totalInFile

            printfn "--------------------------------------------------------------------"
            
            ( (if updateMeta then noSelected else ref -1),
              ref (new StreamReader(File.OpenRead(filename_x))),
              ref (new StreamReader(File.OpenRead(filename_y))),
              ref (new StreamReader(File.OpenRead(filename_filter))))

        
        
    /// Duplicate labels for "dupLabel" times and duplicate features for "dupFeature" times
    /// Not just duplication, but append X1X_, X2X_ ... ahead of the labels, and Y1Y_, Y2Y_ ... ahead of features
    /// This function was written to diagnose how the time and space complexity scales with the data (#feature,#class)
    /// No longer in use after ensuring the implementation is correct with expected complexity.
    let quickDuplicate (dupLabel:int) (dupFeature:int) (line:string)=
    
        let tokens = line.Split [| ' ' |] |> Array.map (fun s -> s.Trim()) |> Array.filter (fun str -> str.Length > 0)
        
        // First blow up the label string
        
        let mutable resultLabelString = tokens.[0]
        let labels = resultLabelString.Split [| ',' |]
        for i in 1 .. (dupLabel - 1) do
            let padString = sprintf ",X%dX_" i
            let newLabels = labels |> Array.map (fun elem -> padString + elem) |> String.Concat
            resultLabelString <- resultLabelString + newLabels
            
        let mutable resultFeatureString = ""
        let noActiveFeature = tokens.Length - 1
        let features = tokens.[1..] 
        for i in 0 .. (dupFeature - 1) do            
            let padString = if i = 0 then " " else sprintf " Y%dY" i
            let newFeatures = features |> Array.map (fun elem -> padString + elem) |> String.Concat
            resultFeatureString <- resultFeatureString + newFeatures

        resultLabelString + resultFeatureString
        
    
    /// Make some statistics accessible    
    member this.NoInstance = 
                if !noInstance < 0 then
                
                    this.rewind () 
                    noInstance := 0
                    
                    let line_skip = ref 0
                    while not (!reader_filter).EndOfStream do

                        totalInFile := !totalInFile + 1
                        if selToken = Int32.of_string ((!reader_filter).ReadLine()) then

                            while not (!reader_x).EndOfStream && not (!reader_y).EndOfStream && !line_skip <> 0 do                            
                                    line_skip := !line_skip - 1
                                    (!reader_x).ReadLine() |> ignore
                                    (!reader_y).ReadLine() |> ignore
                            
                            if !line_skip = 0 then
                                if not (!reader_x).EndOfStream && not (!reader_y).EndOfStream then
                                    (!reader_x).ReadLine() |> ignore
                                    (!reader_y).ReadLine() |> ignore                                    
                                    noInstance := !noInstance + 1
                                else
                                    print_warning("Filter file is longer than feature/label data file");
                        else
                            line_skip := !line_skip + 1
                    
                    totalInFile := !totalInFile - !line_skip

                    while not (!reader_y).EndOfStream do
                    
                        (!reader_y).ReadLine() |> ignore
                        totalInFile := !totalInFile + 1
                    
                !noInstance     
                    
                    
    member this.increNoInstance () = noInstance := !noInstance + 1
    member this.clearnstance () = noInstance := 0
    
    member this.totalInDataset with get() = !totalInFile
    member this.numPos with get() = _numPos
    member this.CurInstance with get () = !curInstance
    
    member this.getNextInstance () = 

        /// When openning, we have made sure that the index file is longer than the train/test file
        if updateMeta && !curInstance < !noInstance || not updateMeta then
            while not ((!reader_x).EndOfStream || (!reader_y).EndOfStream) 
                  && selToken <> Int32.of_string ((!reader_filter).ReadLine()) do
                (!reader_x).ReadLine() |> ignore
                (!reader_y).ReadLine() |> ignore

            if (!reader_x).EndOfStream || (!reader_y).EndOfStream then
                None
            else 
                curInstance := !curInstance + 1 
                
                let line_y = (!reader_y).ReadLine()
                let line_x = 
                    if modelParam.useBias then
                        (!reader_x).ReadLine() + " b:" + (Float.to_string modelParam.biasValue)
                    else
                        (!reader_x).ReadLine() 

                Some(new DataPointSparse<'a> (line_y, line_x, normalize, metaData, modelParam))
        else
            None
       
    /// Rewind the file pointer
    member this.rewind () =    
        
//        printfn "Rewinding %s" filename_x
//        printfn "Rewinding %s" filename_y
       
        (!reader_x).Close()
        reader_x := new StreamReader(File.OpenRead(filename_x))
        
        (!reader_y).Close()
        reader_y := new StreamReader(File.OpenRead(filename_y))
     
        (!reader_filter).Close()
        reader_filter := new StreamReader(File.OpenRead(filename_filter))
        
        curInstance := 0

    /// Close both feature and label files
    member this.Close () = curInstance := 0; [| !reader_x; !reader_y; !reader_filter |] |> Seq.iter (fun x -> x.Close())
    

    /// Print the whole dataset
    /// Need meta-data to convert id to original names of feature and label
    member this.print () = 
        printfn "Labels from %s\nfeatures from %s\nfilter from %s" filename_y filename_x filename_filter
        printfn "Over all data points, there are %d labels and %d features." metaData.NoLabel metaData.NoFeature
        
        printfn "num positive = "
        _numPos |> Seq.iter (fun kvp -> printfn "%s:%A" (metaData.LabelIdToName).[kvp.Key] kvp.Value)
        printfn "--------------------------------------------------------------------"

        [1..!noInstance] |> Seq.iter (fun i ->
            match this.getNextInstance () with
            | None -> failwith "oops, bug in Datasets"
            | Some(dataPoint) -> dataPoint.print(metaData.LabelIdToName, metaData.FeatureIdToName))
            

    member this.printLabelSummary () = 

        printfn "num positive = "
        _numPos |> Seq.iter (fun kvp -> printfn "%s:%A" (metaData.LabelIdToName).[kvp.Key] kvp.Value)
        printfn "--------------------------------------------------------------------"
            
            
            
            
            
//    // save the dataset by sorting the id of feature and label
//    // Also save the dictionaries          
//    member this.sortAndSaveToFile () =
//        let labelIdToName = metaData.LabelIdToName
//        let featureIdToName = metaData.FeatureIdToName
//
//        // first save the dictionary of feature and labels
//        let dot_position =  String.rindex filename_x '.'
//        let (labelDictionaryFilename, featureDictionaryFilename, labelFilename, featureFilename) = 
//            if dot_position = 0 then
//                (filename_x + "_labels.dat", filename_x + "_features.dat", filename_x + "_sorted_X.dat", filename_x + "_sorted_Y.dat")
//            else
//                let head = String.sub filename_x 0 (dot_position-1)
//                let tail = String.sub filename_x (dot_position + 1) (filename_x.Length - dot_position - 1)
//                (head + "labels." + tail, head + "features." + tail, head + "sorted_X." + tail, head + "sorted_Y." + tail)
//                            
//        // Write the label dictionary                                            
//        let labelWriter = new StreamWriter (File.OpenWrite(labelDictionaryFilename))
//        labelIdToName |> Seq.iter (fun kvp -> labelWriter.WriteLine (sprintf "%d\t%s" kvp.Key kvp.Value))
//        labelWriter.Close()
//
//        // Write the feature dictionary
//        let featureWriter = new StreamWriter (File.OpenWrite(featureDictionaryFilename))
//        featureIdToName |> Seq.iter (fun kvp -> featureWriter.WriteLine (sprintf "%d\t%s" kvp.Key kvp.Value))       
//        featureWriter.Close()
//                          
//        // Write the real data
//        let read_x = new StreamReader(File.OpenRead(filename_x))
//        let read_y = new StreamReader(File.OpenRead(filename_y))
//
//        let labelWriter = new StreamWriter(File.OpenWrite(labelFilename))
//        let featureWriter = new StreamWriter(File.OpenWrite(featureFilename))
//        
//        let noLoaded = ref 0
//        while not read_x.EndOfStream do
//            let datapoint = new DataPointSparse(read_y.ReadLine(), read_x.ReadLine(), normalize, metaData)
//            datapoint.sortAndSaveToFile (labelWriter, featureWriter)
//            noLoaded := !noLoaded + 1
//                
//        read_x.Close ()
//        read_y.Close ()
//        labelWriter.Close()  
//        featureWriter.Close()  
//
//       
//   
//    // randomly shuffle the data points and save into file
//    // If we don't allow random access of instances, there is no quick way to randomize
//    member this.randomizeAndSave (seed:int) =
//
//        // first save the dictionary of feature and labels
//        let dot_position =  String.rindex filename_x '.'
//        let (newLabelFilename, newFeatureFilename) = 
//            if dot_position = 0 then
//                (filename_y + "_rand", filename_x + "_rand")
//            else
//                let head = String.sub filename_x 0 (dot_position - 1)
//                let tail = String.sub filename_x (dot_position + 1) (filename_x.Length - dot_position - 1)
//                (head + "rand_Y." + tail, head + "rand_X." + tail)
//                         
//        let labelWriter = new StreamWriter(File.OpenWrite(newLabelFilename))
//        let featureWriter = new StreamWriter(File.OpenWrite(newFeatureFilename))
//        
//        let rnd = System.Random(seed)
//
//        // from (random) value to index
//        let dictionary = new Dictionary<int, int> ()
//                
//        let randomValueList = List.init this.NoInstance (fun i ->
//            let mutable randValue = rnd.Next()
//            while dictionary.ContainsKey randValue do
//                randValue <- rnd.Next()
//            dictionary.Add (randValue, i)
//            randValue)
//        
//        let sortedValueList = randomValueList |> List.sort (fun a b -> a - b)
//
//        sortedValueList |> List.iteri (fun i value ->
//                if i % 200 = 0 then
//                    printfn "%d done" i
//                let idx = dictionary.[value] 
//                ()
//// This line is broken now                
////                dataSet.[idx].sortAndSaveToFile(labelWriter, featureWriter)
//            )
//
//        labelWriter.Close()              
//        featureWriter.Close()  


type DataBundle<[<Measure>] 'a> (_root:string, _datasetName:string, _filterFname:string, normalizeFeatures:bool, 
                                 debugInfo:DebugParam, modelParam : ModelParameter<'a> ) =
        
    let dataDirectory = _root + @"/Dataset/" + _datasetName
    let prefix = dataDirectory + @"/" + _datasetName + "_"
    
    let filename_dataset_x = prefix  + "X.dat"
    let filename_dataset_y = prefix + "Y.dat"
    let filename_filter = dataDirectory + @"/filters/" + _datasetName + "_" + _filterFname + @".dat"


    let _metaData = new MetaData (debugInfo)

    let _trainData = new SeqDataSet <'a>  (filename_dataset_x, filename_dataset_y, filename_filter, 
                                        normalizeFeatures, true, 1, _metaData, modelParam)
    let _testData = new SeqDataSet <'a>  (filename_dataset_x, filename_dataset_y, filename_filter, 
                                        normalizeFeatures, false, 2, _metaData, modelParam)
    let _validData = new SeqDataSet <'a>  (filename_dataset_x, filename_dataset_y, filename_filter, 
                                        normalizeFeatures, true, 3, _metaData, modelParam)    

    member this.trainData with get () = _trainData
    member this.testData with get () = _testData
    member this.validData with get () = _validData
    
    member this.metaData with get () = _metaData
    
    member this.datasetName with get () = _datasetName
    member this.filterFname with get () = _filterFname
    member this.root with get () = _root
    
    member this.print () = _metaData.print ()
            
        
type LearningTask <[<Measure>] 'a>  (_dataBundle : DataBundle<'a>, _modelParam : ModelParameter<'a>, _debugInfo : DebugParam) = 
    
    do _modelParam.updateMonitorInterval _dataBundle.trainData.NoInstance
    
    member this.dataBundle with get () = _dataBundle
    member this.modelParam with get () = _modelParam
    member this.debugInfo with get () = _debugInfo
    
    member this.print () =
        _dataBundle.print ()
        _modelParam.print ()
        _debugInfo.print ()
        _dataBundle.trainData.printLabelSummary ()
                
        if _debugInfo.printTrainData then _dataBundle.trainData.print ()
        if _debugInfo.printTestData then _dataBundle.testData.print ()


type MinHeap (numElem:int) = 
    
    let buffer = Array.create numElem 0.0
    let numAdded = ref 0
       
    member this.add (value:float) =
        
        numAdded := !numAdded + 1
        
        if !numAdded <= numElem then 
            buffer.[!numAdded - 1] <- value
            if !numAdded = numElem then
                buffer |> Array.sort (fun a b -> if a > b then 1 elif a < b then -1 else 0)
        else            
            if value > buffer.[0] then
                let (parent, left_son, right_son) = (ref 0, ref 1, ref 2)
                            
                let violate = ref ((!left_son < numElem && buffer.[!left_son] < value) 
                                    || (!right_son < numElem && buffer.[!right_son] < value))

                while !violate do

                    if !right_son < numElem && buffer.[!right_son] < buffer.[!left_son] then
                        buffer.[!parent] <- buffer.[!right_son]
                        parent := !right_son
                    else
                        buffer.[!parent] <- buffer.[!left_son]
                        parent := !left_son

                    left_son := !parent * 2 + 1
                    right_son := !parent * 2 + 2
                    
                    violate := (!left_son < numElem && buffer.[!left_son] < value) 
                                    || (!right_son < numElem && buffer.[!right_son] < value)
                                    
                buffer.[!parent] <- value

    member this.getTop () = 
        buffer.[0]


