//--------------------------------------------------------------
// gaussians.fs           Gaussian variables and factors
//
// 2007/2008 written by Ralf Herbrich
// Microsoft Research Ltd
//--------------------------------------------------------------

#light

/// small change

namespace MSRC.Inference.Distributions

open System
open System.IO
open MSRC.Inference



// A Gaussian distribution based on float numbers (struct type for memory efficency) 
// in exponential parameterisation. 
type Gaussian<[<Measure>] 'a> = 
    struct
        /// Precision times the mean of the Gaussian
        val PrecisionMean   : float<1/'a> 
        /// Precision of the Gaussian
        val Precision       : float<1/'a^2> 
//        val name : string 
    end
    with
        /// Constructor using the precision times mean and the precision
        new (pm, p) = { PrecisionMean = pm; Precision = p;} 
//        new (pm, p, arg_name) = { PrecisionMean = pm; Precision = p; name = arg_name;} 
        new (g : Gaussian<'a>) = { PrecisionMean = g.Pm; Precision = g.Pre; } 

        member this.Pm = this.PrecisionMean // with get () = this.Pm and set arg_pm = this.PrecisionMean <- arg_pm
        member this.Pre = this.Precision    // with get () = this.Precision and set arg_pre = this.Precision <- arg_pre
        
        // Mean of the Gaussian
        member this.Mu = this.PrecisionMean / this.Precision
        /// Mean of the Gaussian
        member this.Mean = this.Mu
        /// Variance of the Gaussian
        member this.Variance : float<'a^2> = 1.0 / this.Precision
        /// Standard deviation of the Gaussian
        member this.StandardDeviation = sqrt (this.Variance)
        /// Standard deviation of the Gaussian
        member this.Sigma = this.StandardDeviation
        
        member this.Name = "" //with get () = this.name and set arg_name = this.name <- arg_name

//        member this.resetValue (gau : Gaussian<'a>) =
//                this.PrecisionMean <- gau.Pm
//                this.Precision <- gau.Pre

        /// Creates a Gaussian in (mean,standard-deviation) coordinates
        static member Create (mu:float<'a>, sigma:float<'a>) = 
            let sigmaSquared = sigma * sigma
            new Gaussian<'a> (mu / sigmaSquared, 1.0 / sigmaSquared)

        /// Multiplies two Gaussians  
        static member ( * ) (a:Gaussian<'a>, b:Gaussian<'a>) = 
            new Gaussian<'a> (a.PrecisionMean + b.PrecisionMean, a.Precision + b.Precision)

        static member ( / ) (a:Gaussian<'a>, b:Gaussian<'a>) = 
            new Gaussian<'a> (a.PrecisionMean - b.PrecisionMean, a.Precision - b.Precision)

        /// Computes the absolute difference between two Gaussians
        static member AbsoluteDifference (a:Gaussian<'a>, b:Gaussian<'a>) = 
//             max (abs (a.PrecisionMean - b.PrecisionMean)) (sqrt (abs (a.Precision - b.Precision)))
             // let's use relative difference
            let temp2 = sqrt b.Precision
            let a1 = 
                if (abs b.PrecisionMean) > 1e5<_> then 
                    (abs (a.PrecisionMean - b.PrecisionMean) / b.PrecisionMean) * 1.0<_>
                else 
                    abs (a.PrecisionMean - b.PrecisionMean)
            let a2 = 
                if (abs b.Precision) > 1e5<_> then 
                    (sqrt (abs (a.Precision - b.Precision) / b.Precision)) * 1.0<_>
                else 
                   sqrt (abs (a.Precision - b.Precision))

            max a1 a2

        /// Computes the absolute difference between two Gaussians
        static member (-) (a:Gaussian<'a>, b:Gaussian<'a>) = 
            Gaussian<'a>.AbsoluteDifference (a, b)
            
        static member print (a:Gaussian<'a>) =
            printfn "Mu = %A, Sigma = %A, PrecisionMean = %A, Precision = %A" a.Mu a.StandardDeviation a.PrecisionMean a.Precision
            
        static member LogProductNormalisation (a:Gaussian<'a>, b:Gaussian<'a>) =             
            if a.Precision = 0.0<_> then 
                0.0
            elif b.Precision = 0.0<_> then
                0.0
            else                
                let varSum = a.Variance + b.Variance
                let muDiff = a.Mean - b.Mean
                let result = -0.91893853320467267 - 0.5 * log (float varSum) - muDiff * muDiff / (2.0 * varSum)
//                printf "|ha %g| " result

                if Double.IsNaN result then
                    printfn "isnan in LogProductNormalisation: %A, %A" varSum muDiff
                result
                
        /// Computes the log-normalisation factor when two normalised Gaussians gets divided
        static member LogRatioNormalisation (g1:Gaussian<'a>, g2:Gaussian<'a>) =
            if g1.Precision = 0.0<_> then 
                0.0
            elif g2.Precision = 0.0<_> then
                0.0
            else
                let v2 = g2.Variance
                let varDiff = v2 - g1.Variance
                let muDiff = g1.Mean - g2.Mean
                if varDiff = 0.0<_> then
                    0.0
                else
                    let result = log(float v2) + 0.91893853320467267 
                                    - log(float varDiff) / 2.0 + muDiff * muDiff / (2.0 * varDiff)
                    if Double.IsNaN result then
                        printfn "isnan in LogRatioNormalisation: %A, %A" varDiff muDiff
                        
                    result


            
module Gaussians =
    /// Computes the complementary error function. This function is defined     
    /// by 2/sqrt(pi) * integral from x to infinity of exp (-t^2) dt

    // for printing debug information
    let mutable debug = false
    let mutable show_factor = false
    
    // For message capping in order to preserve message accuracy
    let maxPi = 1e10</Weight^2>
    let minPi = 0.0</Weight^2>
    
    /// all messages that have precision less than zeroPrecisionTolerance are deems as 0 precision
    /// used in the weighted-sum factors (or its fancy variants)
    let zeroPrecisionTolerance = 1e-10</Weight^2>
    
    /// thresholds used in the Greater Than factor
    let t_search_tol = 0.001  // precision of the search for t_low
    let v_cdf_threshold = 2.222758749e-300
    let w_eps_low = 1e-10
    // we currently do not cap when the input to the GreaterThan factor is very large.  Maybe in the future, we will cap as well.
    let w_eps_high = 0.0    
    
    /// whether to check the latest training point is immediately fit
    let check_immediate_fit = false
    let printVarName = false
    
    /// Decay parameters
    let newSampleWeight = 1.0
    let decayRate = 0.0  // current decay factor is newSampleWeight / ( 1 + decayRate * t )
    
    let carelabelOrderToId = false

    let erfc x =
        if (System.Double.IsNegativeInfinity (x)) then
            2.0
        else if (System.Double.IsPositiveInfinity (x)) then
            0.0
        else
            let z = abs (x) 
            let t = 1.0 / (1.0 + 0.5 * z) 
            let res = t * exp (-z * z - 1.26551223 + t * (1.00002368 + t * (0.37409196 + t * (0.09678418 + t * (-0.18628806 + t * (0.27886807 + t * (-1.13520398 + t * (1.48851587 + t * (-0.82215223 + t * 0.17087277))))))))) 
            if (x >= 0.0) then res else 2.0 - res

    /// Computes the inverse of the complementary error function
    let erfcinv y = 
        if (y < 0.0 || y > 2.0) then
            failwith "Inverse complementary function not defined outside [0,2]."
        elif (y = 0.0) then
            System.Double.PositiveInfinity
        elif (y = 2.0) then
            System.Double.NegativeInfinity
        else 
            let x = 
                if (y >= 0.0485 && y <= 1.9515) then
                    let q = y - 1.0 
                    let r = q * q 
                    (((((0.01370600482778535*r - 0.3051415712357203)*r + 1.524304069216834)*r - 3.057303267970988)*r + 2.710410832036097)*r - 0.8862269264526915) * q /
                    (((((-0.05319931523264068*r + 0.6311946752267222)*r - 2.432796560310728)*r + 4.175081992982483)*r - 3.320170388221430)*r + 1.0)
                else if (y < 0.0485) then
                    let q = sqrt (-2.0 * log (y / 2.0)) 
                    (((((0.005504751339936943*q + 0.2279687217114118)*q + 1.697592457770869)*q + 1.802933168781950)*q + -3.093354679843504)*q - 2.077595676404383) / 
                    ((((0.007784695709041462*q + 0.3224671290700398)*q + 2.445134137142996)*q + 3.754408661907416)*q + 1.0)
                else if (y > 1.9515) then
                    let q = sqrt (-2.0 * log (1.0 - y / 2.0)) 
                    (-(((((0.005504751339936943*q + 0.2279687217114118)*q + 1.697592457770869)*q + 1.802933168781950)*q + -3.093354679843504)*q - 2.077595676404383) / 
                     ((((0.007784695709041462*q + 0.3224671290700398)*q + 2.445134137142996)*q + 3.754408661907416)*q + 1.0))
                else 0.0
            let u = (erfc (x) - y) / (-2.0 / sqrt (System.Math.PI) * exp (-x * x)) 
            x - u / (1.0 + x * u)

    /// Computes the cummulative Gaussian distribution at a specified point of interest
    let normcdf t = 
        let sqrt2 = 1.4142135623730951 
        (erfc (-t / sqrt2)) / 2.0
        
    let lognormcdf t =
        if t > -5.0 then log(normcdf t) else 
            // using approximation from http://eprints.lancs.ac.uk/596/2/7BF32BA0.pdf, with a = 3
            let a = 1.5707963267948966 * t * t  
            0.40546510810816438 - t * t / 2.0 - log (2.0 * sqrt(a) + sqrt(a + 9.0))

    /// Computes the Gaussian density at a specified point of interest
    let normpdf (t:float) = 
        let invsqrt2pi = 0.398942280401433
        invsqrt2pi * exp (- (t * t / 2.0))
        
    /// Computes the inverse of the cummulative Gaussian distribution (qunatile function) at a specified point of interest
    let norminv p = 
        let sqrt2 = 1.4142135623730951 
        (-sqrt2 * erfcinv (2.0 * p))

    /// Computes the cummulative Gaussian distribution at a specified point of interest
    let Phi = normcdf
        
    /// Computes the inverse of the cummulative Gaussian distribution (qunatile function) at a specified point of interest
    let PhiInverse = norminv


    /// Computes the additive correction of a single-sided truncated Gaussian with unit variance
    // when it comes here, it is assumed that t_low < t
    let v t epsilon = 
//        match normcdf (t-epsilon) with
//        | denom when denom < 2.222758749e-162   -> printfn "inside" ; -t+epsilon
//        | denom                                 -> (normpdf (t - epsilon)) / denom                             
          normpdf (t - epsilon) / normcdf (t - epsilon)

    /// Computes the multiplicative correction of a single-sided truncated Gaussian with unit variance
    let w t epsilon =
//        match normcdf (t-epsilon) with
        //| denom when denom < 2.222758749e-162   -> if (t < 0.0) then 1.0 else 0.0
        //| denom when denom < 2.222758749e-162   -> printfn "capping===================================================================="; if (t < 0.0) then 1.0 - small_eps else small_eps
        
          let temp = v t epsilon
          temp * (temp + t - epsilon)


    /// Computes the additive correction of a double-sided truncated Gaussian with unit variance
    let v0 t epsilon = 
        let v = abs t
        match normcdf (epsilon - v) - normcdf (-epsilon - v) with
        | denom when denom < 2.222758749e-162   -> if t < 0.0 then -t-epsilon else -t+epsilon
        | denom                                 -> let num = normpdf (-epsilon-v) - normpdf (epsilon-v) in if t < 0.0 then -num/denom else num/denom

    /// Computes the multiplicative correction of a double-sided truncated Gaussian with unit variance
    let w0 t epsilon =        
        let v = abs t
        match normcdf (epsilon - v) - normcdf (-epsilon - v) with
        | denom when denom < 2.222758749e-162   -> 1.0
        | denom                                 -> let vt = v0 v epsilon in vt*vt + ((epsilon-v) * normpdf (epsilon-v) - (-epsilon-v) * normpdf (-epsilon-v))/denom


    /// find the truncation points fo t, and their corresponding v and w values
    let (t_low, t_high, v_low, v_high, w_low, w_high) = 
    
        // check whether cur_t satisfies that w(cur_t) > 1 - w_eps_low
        let check_lower cur_t =
            let mutable temp_v = normcdf cur_t            
            if temp_v < v_cdf_threshold then
                true
            else
                temp_v <- (normpdf cur_t) / temp_v
                let temp_w = temp_v * (temp_v + cur_t)
                if temp_w < 1.0 - w_eps_low then
                    false
                else
                    true

        // check whether cur_t satisfies that w(cur_t) < w_eps
        let check_higher cur_t =
            let mutable temp_v = normcdf cur_t
            if temp_v < v_cdf_threshold then    // this should actually never happen if the starting point is not too crazy
                false
            else
                temp_v <- (normpdf cur_t) / (normcdf cur_t)
                let temp_w = temp_v * (temp_v + cur_t)
                if temp_w < w_eps_high then
                    true
                else
                    false

        let mutable cur_low = -1.0
        let mutable cur_high = 1.0
        let mutable cur_mid = 0.0
        let mutable temp_w = 0.0
        let mutable temp_v = 0.0
                  
        /////////////////////////////////////////////
        // first find the t_low
        while not (check_lower cur_low) do
            cur_low <- cur_low * 2.0
        cur_high <- cur_low / 2.0
        
        cur_mid <- cur_low        
        while (cur_high - cur_low) > t_search_tol do            
            cur_mid <- (cur_high + cur_low) / 2.0
            if check_lower cur_mid then
                cur_low<- cur_mid
            else
                cur_high <- cur_mid
                                
        let t_low = cur_mid
        let v_low = (normpdf t_low) / (normcdf t_low)
        let w_low = v_low * (v_low + t_low)
        
        /////////////////////////////////////////////
        let (t_high, v_high, w_high) = 
        
            if w_eps_high = 0.0 then
                (Double.MaxValue, 0.0, 0.0)
            else    
                cur_high <- 1.0
                
                // first find the t_high
                while not (check_higher cur_high) do
                    cur_high <- cur_high * 2.0
                cur_low <- cur_high / 2.0
                
                cur_mid <- cur_high
                while cur_high - cur_low > t_search_tol do
                    
                    cur_mid <- (cur_high + cur_low) / 2.0
                    if check_higher cur_mid then
                        cur_high <- cur_mid
                    else
                        cur_low <- cur_mid
                let temp_vhigh = (normpdf cur_mid) / (normcdf cur_mid)
            
                (cur_mid, temp_vhigh, temp_vhigh * (temp_vhigh + cur_mid))
        
        // Print out some information
        printfn "t_low = %g, t_high = %g, v_low = %g, v_high = %g, w_low = %g, w_high = %g" t_low t_high v_low v_high w_low w_high
    
        (t_low * 1.0<_>, t_high * 1.0<_>, v_low, v_high, w_low, w_high)




namespace MSRC.Inference

open MSRC.Inference.Collections
open MSRC.Inference.Factors
open MSRC.Inference.Distributions
open MSRC.Inference.Distributions.Gaussians
open System


type GaussianFactors<[<Measure>] 'a> () = 
    /// Defines a bag of distributions that holds all the Gaussian message 
    let MessageBag = new DistributionBag<Gaussian<'a>> (new Gaussian<'a> (0.0<_>, 0.0<_>))
    
    /// Allocates a new Gausian with parameter Mean=mu and Variance=sigma2
    let GaussianMS (mu:float<'a>, sigma2:float<'a^2>) = new Gaussian<'a> (mu / sigma2, 1.0 / sigma2)

    /// Sends the message witha given index to a variable and returns the log-normalisation constant
    let SendMessageHelper msgIdx (varIdx, varBag:DistributionBag<Gaussian<'a>>) =         
        let mar = varBag.[varIdx]
        let msg = MessageBag.[msgIdx] 
        let logZ = Gaussian.LogProductNormalisation (mar, msg)
        varBag.[varIdx] <- mar * msg
//        printfn "sm %A" logZ
        logZ          
    
    
    /// cap the precision.  
    /// Prototype: (precisionMu, precision) -> (new_precisionMu, new_precision)
    let CappingGaussian (precisionMu:float<1/'a>, precision:float<1/'a^2>) = 
        if (precision <= 1.0<_> / System.Double.MaxValue) then   // in case of (0, 0), just keep (0, 0) 
          (precisionMu, precision)
        elif precision <= maxPi * 1.0<_> && precision >= minPi * 1.0<_> then
            (precisionMu, precision)
        elif precision > maxPi * 1.0<_> then
            if debug && show_factor then
                printfn "using capping from above, before: %A, %A" precisionMu precision
            (precisionMu / precision * (maxPi * 1.0<_>), maxPi * 1.0<_>)
        else
            // This should never occur, when precision is small, just let it small.
            if debug && show_factor then
                printfn "using capping from below, before: %A, %A, after: %A, %A, min=%A, max=%A" precisionMu precision (precisionMu / precision * minPi) minPi
                                                                                minPi maxPi
            (precisionMu / precision * (minPi * 1.0<_>), minPi * 1.0<_>)   
                
        
    let check_ration value message =
        
        if Double.IsNaN value then printfn "is nan in %s" message
        if Double.IsPositiveInfinity value then printfn "is pos_infinity in %s" message
        if Double.IsNegativeInfinity value then printfn "is neg_infinity in %s" message



    // (tau, pi, mu, sigma^2)
    let GreaterThanStat epsilon (gau:Gaussian<'a>) =

        let c = gau.Precision
        let d = gau.PrecisionMean
        let sqrtC = sqrt c
        let dOnSqrtC = d / sqrtC
        let epsTimesSqrtC = epsilon * sqrtC
        let tMinusEps = dOnSqrtC - epsTimesSqrtC
                           
        // compute the message (not the new marginal!)
        let (v_val, w_val) = 

            if tMinusEps > t_low then
                ( v dOnSqrtC epsTimesSqrtC, w dOnSqrtC epsTimesSqrtC)
            else
                //let v_val = (sqrt (tMinusEps * tMinusEps + 4.0 * w_val) - tMinusEps) / 2.0
                ( (-tMinusEps + sqrt(tMinusEps * tMinusEps + 4.0 * w_low)) / 2.0, w_low)

        let (newTau, newPi) = CappingGaussian ((sqrtC * v_val + d * w_val) / ( 1.0 - w_val), w_val * c / (1.0 - w_val))
        (newTau, newPi, newTau / newPi, 1.0 / newPi)



    // (tau, pi, mu, sigma^2)
    let LessThanStat epsilon (gau:Gaussian<'a>) =

        let c = gau.Precision
        let d = - gau.PrecisionMean
        let sqrtC = sqrt c
        let dOnSqrtC = d / sqrtC
        let epsTimesSqrtC = epsilon * sqrtC
        let tMinusEps = dOnSqrtC - epsTimesSqrtC
                           
        // compute the message (not the new marginal!)
        let (v_val, w_val) = 
            if tMinusEps > t_low then
                (v dOnSqrtC epsTimesSqrtC, w dOnSqrtC epsTimesSqrtC)
            else
                //let v_val = (sqrt (tMinusEps * tMinusEps + 4.0 * w_val) - tMinusEps) / 2.0
                ((-tMinusEps + sqrt(tMinusEps * tMinusEps + 4.0 * w_low)) / 2.0, w_low)

        let (newTau, newPi) = CappingGaussian ((sqrtC * v_val + d * w_val) / ( 1.0 - w_val), w_val * c / (1.0 - w_val))
        ( -newTau, newPi, - newTau / newPi, 1.0 / newPi )



    /// Removes all messages stored so far
    member this.ClearMessages () = MessageBag.Reset ()    
        
    
    member this.ClearSomeMessages (noToKeep) = 
        MessageBag.truncate(noToKeep)
        

    /// Connects a Gaussian priot factor to variable and returns the update function
    member this.GaussianPriorFactor (mu,sigma2) (varIdx, varBag:DistributionBag<Gaussian<'a>>) name = 
        let msgIdx = MessageBag.NewDistribution ()
        let newMsg = ref (GaussianMS (mu, sigma2))
       
        { 
            // ChangePrior allows us to do ADF efficiently
            // no need to rellocate space or update the message/variable lists
            ChangePrior = fun (pm, p) ->
                newMsg := new Gaussian<'a> (pm, p)
            
            getPrior = fun () ->
                ((!newMsg).PrecisionMean, (!newMsg).Precision)

            Name = fun () -> name
            
            getType = fun () -> 0

            UpdateMessage = fun idx ->
                match idx with
                | 0 -> 
                    let oldMarginal = varBag.[varIdx]
                    let oldMsg = MessageBag.[msgIdx]
                    let (newMarPreMu, newMarPre) = CappingGaussian (oldMarginal.PrecisionMean + (!newMsg).PrecisionMean - oldMsg.PrecisionMean, 
                                                        oldMarginal.Precision + (!newMsg).Precision - oldMsg.Precision)
                    let newMarginal = new Gaussian<'a> (newMarPreMu, newMarPre) //, oldMarginal.Name
                    varBag.[varIdx] <- newMarginal
                    MessageBag.[msgIdx] <- (!newMsg)
                    
                    if debug && show_factor then
                        printfn "In GaussianPriorFactor, Msg %s -> %s: (%A, %A), new marginal = (%A, %A)" name varBag.[varIdx].Name
                            (!newMsg).PrecisionMean (!newMsg).Precision newMarginal.PrecisionMean newMarginal.Precision
                    oldMarginal - newMarginal
                    
                | _ -> failwith "Variable index out of range. Valid values are 0 only."
            LogNormalisation = fun () -> 0.0
            NumberOfMessages = 1
            ResetMarginals = fun () -> varBag.[varIdx] <- varBag.Prior
            SendMessage = fun idx ->
                match idx with
                | 0 -> SendMessageHelper msgIdx (varIdx,varBag) 
                | _ -> failwith "Variable index out of range. Valid values are 0 only."
        }

    /// Connects a Gaussian likelihood factor to 2 variables and returns the update function
    /// The distributions are related by N(var0;var1,betaSquared).
    member this.GaussianLikelihoodFactor (betaSquared:float<'a^2>) (var1Idx, var1Bag:DistributionBag<Gaussian<'a>>) (var2Idx,var2Bag:DistributionBag<Gaussian<'a>>) name = 
        let msg1Idx = MessageBag.NewDistribution ()
        let msg2Idx = MessageBag.NewDistribution ()
        let prec = 1.0 / betaSquared 
        
        // send from v2 to v1
        let UpdateHelper m1 m2 (v1,b1:DistributionBag<Gaussian<'a>>) (v2,b2:DistributionBag<Gaussian<'a>>) = 
            let msg1 = MessageBag.[m1]
            let msg2 = MessageBag.[m2]
            let mar1 = b1.[v1]
            let mar2 = b2.[v2]
            let a = prec / (prec + mar2.Precision - msg2.Precision)
            let newMsg = Gaussian (a * (mar2.PrecisionMean - msg2.PrecisionMean), a * (mar2.Precision - msg2.Precision))
            let oldMarginalWithoutMsg = mar1 / msg1
            let newMarginal = oldMarginalWithoutMsg * newMsg 
            
//            if printVarName then
//                newMarginal.Name <- mar1.Name
            /// Update the message and marginal
            MessageBag.[m1] <- newMsg
            b1.[v1] <- newMarginal
            
            if debug && show_factor then
                printfn "In GaussianLikelihoodFactor, Msg %s -> %s: (%A, %A), new marginal = (%A, %A)" name (b1.[v1].Name)
                    newMsg.PrecisionMean newMsg.Precision newMarginal.PrecisionMean newMarginal.Precision
                    
            /// Return the difference in the new marginal
            newMarginal - mar1
            
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            
            Name = fun () -> name
            
            getType = fun () -> 1
            
            UpdateMessage = fun idx ->
                match idx with
                | 0 ->  UpdateHelper msg1Idx msg2Idx (var1Idx, var1Bag) (var2Idx, var2Bag)
                | 1 ->  UpdateHelper msg2Idx msg1Idx (var2Idx, var2Bag) (var1Idx, var1Bag)
                | _ ->  failwith "Variable index out of range. Valid values are 0 and 1."
                
            LogNormalisation = fun () -> 
                0.0
//                Gaussian.LogRatioNormalisation (var1Bag.[var1Idx], MessageBag.[msg1Idx])

            NumberOfMessages = 2

            ResetMarginals = fun () ->
                var1Bag.[var1Idx] <- var1Bag.Prior
                var2Bag.[var2Idx] <- var2Bag.Prior

            SendMessage = fun idx -> 
                0.0
                
//                match idx with
//                | 0 -> 
////                       printfn "START 0 ==="
////                       printfn "(%A,%A,%A) (%A,%A,%A)" MessageBag.[msg1Idx].PrecisionMean MessageBag.[msg1Idx].Precision (MessageBag.[msg1Idx].Precision = 0.0<_>) var1Bag.[var1Idx].PrecisionMean var1Bag.[var1Idx].Precision (var1Bag.[var1Idx].Precision=0.0<_>)
//                       let result1 = SendMessageHelper msg1Idx (var1Idx,var1Bag)
////                       printfn "%A" result1
////                       System.Console.ReadLine() |> ignore
//                       result1
//                | 1 -> 
////                       printfn "START 1 ==="
////                       printfn "(%A,%A,%A) (%A,%A,%A)" MessageBag.[msg2Idx].PrecisionMean MessageBag.[msg2Idx].Precision (MessageBag.[msg2Idx].Precision = 0.0<_>) var2Bag.[var2Idx].PrecisionMean var2Bag.[var2Idx].Precision (var2Bag.[var2Idx].Precision=0.0<_>)                       
//                       let result1 = SendMessageHelper msg2Idx (var2Idx,var2Bag) 
////                       printfn " %A" result1
////                       System.Console.ReadLine() |> ignore
//                       result1
//                | _ ->  failwith "Variable index out of range. Valid values are 0 and 1."
        }  
                  

    /// Connects a Gaussian weighted sum factor to 3 variables and returns the update function
    /// The distributions are related by var0 = a1*var1 + a2*var2
    member this.GaussianWeightedSumFactor a1 a2 (var0Idx,var0Bag:DistributionBag<Gaussian<'a>>) 
                                                (var1Idx,var1Bag:DistributionBag<Gaussian<'a>>) 
                                                (var2Idx,var2Bag:DistributionBag<Gaussian<'a>>) 
                                                name = 
        let msg0Idx = MessageBag.NewDistribution ()
        let msg1Idx = MessageBag.NewDistribution ()
        let msg2Idx = MessageBag.NewDistribution ()
        let weights0 = [| a1; a2 |]
        let weights1 = [| -a2/a1; 1.0/a1 |]
        let weights2 = [| -a1/a2; 1.0/a2 |]
        let weights0Squared = [| weights0.[0]*weights0.[0]; weights0.[1]*weights0.[1] |]
        let weights1Squared = [| weights1.[0]*weights1.[0]; weights1.[1]*weights1.[1] |]
        let weights2Squared = [| weights2.[0]*weights2.[0]; weights2.[1]*weights2.[1] |]
        if a1 |> System.Double.IsNaN || a2 |> System.Double.IsNaN then
            printfn "hehe GaussianWeightedSumFactor %g %g" a1 a2
            
        let UpdateHelper (w:float array) (wS:float array) m1 m2 m3 (v1,b1:DistributionBag<Gaussian<'a>>) (v2,b2:DistributionBag<Gaussian<'a>>) (v3,b3:DistributionBag<Gaussian<'a>>) = 
            let d0 = b2.[v2] / MessageBag.[m2]
            let d1 = b3.[v3] / MessageBag.[m3]
            let msg1 = MessageBag.[m1]
            let mar1 = b1.[v1]
            let denom = wS.[0] * d1.Precision + wS.[1] * d0.Precision
            
            // First compute the new message
            let (newPrecisionMean, newPrecision) =           
                if  (min d1.Precision d0.Precision) * 1.0<_> > zeroPrecisionTolerance then    //temp > zeroPrecisionTolerance
                    ((w.[0] * d1.Precision * d0.PrecisionMean + w.[1] * d0.Precision * d1.PrecisionMean) / denom,
                                d0.Precision * d1.Precision / denom)
                else
                    if debug && show_factor then printfn "clamping to 0"
                    (0.0<_>, 0.0<_>)

            let newMsg = Gaussian (newPrecisionMean, newPrecision)
            
            // Second compute the new marginal
            // If all previous messages are in proper range then the marginals should also be in that range.  
            //  Hence no need to cap the marginals explicitly.
            let newMarginal = mar1 / msg1 * newMsg 
            
            /// Update the message and marginal
//            if printVarName then
//                newMarginal.Name <- mar1.name
            MessageBag.[m1] <- newMsg
            b1.[v1] <- newMarginal
            
            if float newPrecision |> System.Double.IsNaN || float newPrecisionMean |> System.Double.IsNaN then
                printfn "d0: %A %A,  d1: %A %A" d0.PrecisionMean d0.Precision d1.PrecisionMean d1.Precision
                printfn "look GaussianWeightedSumFactor %A %A, and %A %A, and denom = %A and original = %A" 
                            (float b2.[v2].PrecisionMean) (float MessageBag.[m2].PrecisionMean) newPrecisionMean newPrecision
                            denom (w.[0] * d1.Precision * d0.PrecisionMean + w.[1] * d0.Precision * d1.PrecisionMean)
                failwith "aaa"
        
            if debug && show_factor then
                printfn "In GaussianWeightedSumFactor, Msg %s -> %s: (%A, %A), new marginal = (%A, %A)" name (b1.[v1].Name)
                    newMsg.PrecisionMean newMsg.Precision newMarginal.PrecisionMean newMarginal.Precision

            /// Return the difference in the new marginal
            newMarginal - mar1
            
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            
            Name = fun () -> name
            
            getType = fun () -> 2
            
            UpdateMessage = fun idx ->
                match idx with
                | 0 ->  UpdateHelper weights0 weights0Squared msg0Idx msg1Idx msg2Idx (var0Idx,var0Bag) (var1Idx,var1Bag) (var2Idx,var2Bag)
                | 1 ->  UpdateHelper weights1 weights1Squared msg1Idx msg2Idx msg0Idx (var1Idx,var1Bag) (var2Idx,var2Bag) (var0Idx,var0Bag)
                | 2 ->  UpdateHelper weights2 weights2Squared msg2Idx msg1Idx msg0Idx (var2Idx,var2Bag) (var1Idx,var1Bag) (var0Idx,var0Bag)
                | _ ->  failwith "Variable index out of range. Valid values are 0, 1 and 2."
                
            LogNormalisation = fun () -> 
                0.0
//                Gaussian.LogRatioNormalisation (var1Bag.[var1Idx], MessageBag.[msg1Idx]) + Gaussian.LogRatioNormalisation (var2Bag.[var2Idx], MessageBag.[msg2Idx])
            
            NumberOfMessages = 3
            
            ResetMarginals = fun () ->
                var0Bag.[var0Idx] <- var0Bag.Prior
                var1Bag.[var1Idx] <- var1Bag.Prior
                var2Bag.[var2Idx] <- var2Bag.Prior
            
            SendMessage = fun idx ->
                0.0
//                match idx with
//                | 0 -> SendMessageHelper msg0Idx (var0Idx,var0Bag) 
//                | 1 -> SendMessageHelper msg1Idx (var1Idx,var1Bag) 
//                | 2 -> SendMessageHelper msg2Idx (var2Idx,var2Bag) 
//                | _ ->  failwith "Variable index out of range. Valid values are 0, 1 and 2."
        }            

    /// Connects a Gaussian greater-than-epsilon factor to variable and returns the update function
    member this.GaussianGreaterThanFactor epsilon (varIdx, varBag:DistributionBag<Gaussian<'a>>) name =
        let msgIdx = MessageBag.NewDistribution ()
        let logEvidence = ref 0.0
        
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            Name = fun () -> name
            getType = fun () -> 3

            UpdateMessage = fun idx ->
                match idx with
                | 0 ->  
                    let oldMarginal = varBag.[varIdx]
                    let oldMsg = MessageBag.[msgIdx]
                    let msgFromVar = oldMarginal / oldMsg

                    
//                    let c = msgFromVar.Precision
//                    let d = msgFromVar.PrecisionMean
//                    let sqrtC = sqrt c
//                    let dOnSqrtC = d / sqrtC
//                    let epsTimesSqrtC = epsilon * sqrtC
//                    let tMinusEps = dOnSqrtC - epsTimesSqrtC
//                    
//                    
//                    // compute the message (not the new marginal!)
//                    let (newPrecisionMean, newPrecision) = 
//                        if tMinusEps > t_low then
//                            let v_val = v dOnSqrtC epsTimesSqrtC
//                            let w_val = w dOnSqrtC epsTimesSqrtC
//                            CappingGaussian((sqrtC * v_val + d * w_val) / ( 1.0 - w_val), w_val * c / (1.0 - w_val))
//                        else
//                            if debug && show_factor then
//                                printfn "******************* using truncation in GaussianGreaterThanFactor ******************* "
//                            let w_val = w_low
//                            //let v_val = (sqrt (tMinusEps * tMinusEps + 4.0 * w_val) - tMinusEps) / 2.0
//                            let v_val = (-tMinusEps + sqrt(tMinusEps * tMinusEps + 4.0 * w_val)) / 2.0
//                            if debug && show_factor then
//                                printfn "check %A\t%A\t%A\t%A" sqrtC v_val d w_val
//                            CappingGaussian ((sqrtC * v_val + d * w_val) / ( 1.0 - w_val), w_val * c / (1.0 - w_val))

                    let (newPrecisionMean, newPrecision, _, _) = GreaterThanStat epsilon msgFromVar
                    let newMsg = Gaussian(newPrecisionMean, newPrecision)
                    
//                    printfn "In GaussianGreaterThanFactor, newMsg = (%A, %A)" newMsg.PrecisionMean newMsg.Precision


                    logEvidence := lognormcdf ((msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)

                    let newMarginal = oldMarginal / oldMsg * newMsg 
//                    if printVarName then
//                        newMarginal.Name <- oldMarginal.Name                  
                    
                    /// Update the message and marginal
                    MessageBag.[msgIdx] <- newMsg
                    varBag.[varIdx] <- newMarginal

//                    if float newMarginal.Precision |> System.Double.IsNaN || float newMarginal.PrecisionMean |> System.Double.IsNaN then
//                        printfn "look GaussianGreaterThanFactor %g, %g, originally %g %g %g and c=%g " (float oldMarginal.PrecisionMean) (float oldMarginal.Precision)
//                                (wVal * d + sqrtC * vVal)  (wVal * c) (1.0 - wVal) c
                    
                    if debug && show_factor then
                        printfn "In GaussianGreaterThanFactor, Msg %s -> %s: (%A, %A), new marginal = (%A, %A)" name (varBag.[varIdx].Name)
                            newMsg.PrecisionMean newMsg.Precision newMarginal.PrecisionMean newMarginal.Precision

                    /// Return the difference in the new marginal
                    
                    newMarginal - oldMarginal
                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
            
            LogNormalisation = fun () -> 
                !logEvidence
//                let marginal = varBag.[varIdx]
//                let msg = MessageBag.[msgIdx]
//                let msgFromVar = marginal / msg
//                let a = -Gaussian.LogProductNormalisation (msgFromVar, msg) 
//                check_ration a "LogProductNormalisation in GaussianGreaterThanFactor"
//                
//                let b = lognormcdf ((msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)
//                check_ration b "log normcdf in GaussianGreaterThanFactor"
//                
//                a + b

            
            NumberOfMessages = 1
            
            ResetMarginals = fun () ->
                varBag.[varIdx] <- varBag.Prior
            
            SendMessage = fun idx ->
                0.0
//                match idx with
//                | 0 -> SendMessageHelper msgIdx (varIdx,varBag) 
//                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
        }            

    /// Connects a Gaussian greater-than-epsilon factor to variable and returns the update function
    member this.GaussianLessThanFactor epsilon (varIdx, varBag:DistributionBag<Gaussian<'a>>) name =
        let msgIdx = MessageBag.NewDistribution ()
        let logEvidence = ref 0.0
        
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            Name = fun () -> name
            getType = fun () -> 4

            UpdateMessage = fun idx ->
                match idx with
                | 0 ->  
                    let oldMarginal = varBag.[varIdx]
                    let oldMsg = MessageBag.[msgIdx]
                    let msgFromVar = oldMarginal / oldMsg
//                    let c = msgFromVar.Precision
//                    let d = - msgFromVar.PrecisionMean  // note the minus
//                    let sqrtC = sqrt c
//                    let dOnSqrtC = d / sqrtC
//                    let epsTimesSqrtC = epsilon * sqrtC
//                    let tMinusEps = dOnSqrtC - epsTimesSqrtC
                    
                    
                    // compute the message (not the new marginal!)
//                    let (newPrecisionMean, newPrecision) = 
//                        if tMinusEps > t_low then
//                            let v_val = v dOnSqrtC epsTimesSqrtC
//                            let w_val = w dOnSqrtC epsTimesSqrtC
//                            CappingGaussian((sqrtC * v_val + d * w_val) / ( 1.0 - w_val), w_val * c / (1.0 - w_val))
//                        else
//                            if debug && show_factor then
//                                printfn "******************* using truncation in GaussianLessThanFactor ******************* "
//                            let w_val = w_low
//                            //let v_val = (sqrt (tMinusEps * tMinusEps + 4.0 * w_val) - tMinusEps) / 2.0
//                            let v_val = (-tMinusEps + sqrt(tMinusEps * tMinusEps + 4.0 * w_val)) / 2.0
//                            if debug && show_factor then
//                                printfn "check %A\t%A\t%A\t%A" sqrtC v_val d w_val
//                            CappingGaussian ((sqrtC * v_val + d * w_val) / ( 1.0 - w_val), w_val * c / (1.0 - w_val))
    
    
                    let (newPrecisionMean, newPrecision, _, _) = LessThanStat epsilon msgFromVar
                    let newMsg = Gaussian(newPrecisionMean, newPrecision)

//                    printfn "In GaussianLessThanFactor, newMsg = (%A, %A)" newMsg.PrecisionMean newMsg.Precision

                    logEvidence := lognormcdf ((-msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)
                    
                    let newMarginal = oldMarginal / oldMsg * newMsg 

//                    if printVarName then
//                        newMarginal.Name <- oldMarginal.Name                  
                    
                    /// Update the message and marginal
                    MessageBag.[msgIdx] <- newMsg
                    varBag.[varIdx] <- newMarginal
                    
//                    if float newMarginal.Precision |> System.Double.IsNaN || float newMarginal.PrecisionMean |> System.Double.IsNaN then
//                        printfn "look GaussianGreaterThanFactor %g, %g, originally %g %g %g and c=%g " (float oldMarginal.PrecisionMean) (float oldMarginal.Precision)
//                                (wVal * d + sqrtC * vVal)  (wVal * c) (1.0 - wVal) c
                    
                    if debug && show_factor then
                        printfn "In GaussianLessThanFactor, Msg %s -> %s: (%A, %A), new marginal = (%A, %A)" name (varBag.[varIdx].Name)
                            newMsg.PrecisionMean newMsg.Precision newMarginal.PrecisionMean newMarginal.Precision

                    /// Return the difference in the new marginal
                    
                    newMarginal - oldMarginal
                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
            
            LogNormalisation = fun () -> 
                !logEvidence
//                let marginal = varBag.[varIdx]
//                let msg = MessageBag.[msgIdx]
//                let msgFromVar = marginal / msg
//
//                let a = -Gaussian.LogProductNormalisation (msgFromVar, msg)
//                check_ration a "LogProductNormalisation of GaussianLessThanFactor"
//                                
//                let b = lognormcdf ((-msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)
//                check_ration b "log normcdf of GaussianLessThanFactor"
//                                
//                a + b
            
            NumberOfMessages = 1
            
            ResetMarginals = fun () ->
                varBag.[varIdx] <- varBag.Prior
            
            SendMessage = fun idx ->
                0.0
//                match idx with
//                | 0 -> SendMessageHelper msgIdx (varIdx,varBag) 
//                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
        }



    /// Connects a Gaussian within-epsilon factor to variable and returns the update function
    member this.GaussianWithinFactor epsilon (varIdx,varBag:DistributionBag<Gaussian<'a>>) name =
        let msgIdx = MessageBag.NewDistribution ()
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            
            Name = fun () -> name
            
            getType = fun () -> 5
            
            UpdateMessage = fun idx ->
                match idx with
                | 0 ->  
                    let oldMarginal = varBag.[varIdx]
                    let oldMsg = MessageBag.[msgIdx]
                    let msgFromVar = oldMarginal / oldMsg
                    let c = msgFromVar.Precision
                    let d = msgFromVar.PrecisionMean
                    let sqrtC = sqrt c
                    let dOnSqrtC = d / sqrtC
                    let epsTimesSqrtC = epsilon * sqrtC
                    let d = msgFromVar.PrecisionMean
                    let denom = 1.0 - w0 dOnSqrtC epsTimesSqrtC
                    let newPrecision = c / denom
                    let newPrecisionMean = (d + sqrtC * v0 dOnSqrtC epsTimesSqrtC) / denom
                    let newMarginal = Gaussian (newPrecisionMean, newPrecision)
                    let newMsg = oldMsg * newMarginal / oldMarginal
                    /// Update the message and marginal
                    MessageBag.[msgIdx] <- newMsg
                    varBag.[varIdx] <- newMarginal
                    /// Return the difference in the new marginal
                    newMarginal - oldMarginal
                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
            
            LogNormalisation = fun () -> 
                let marginal = varBag.[varIdx]
                let msg = MessageBag.[msgIdx]
                let msgFromVar = marginal / msg
                let mean = msgFromVar.Mean
                let std = msgFromVar.StandardDeviation
                let Z = normcdf ((epsilon - mean)/std) - normcdf ((-epsilon-mean)/std)
                -Gaussian.LogProductNormalisation (msgFromVar, msg) + log(Z)
            
            NumberOfMessages = 1
            
            ResetMarginals = fun () ->
                varBag.[varIdx] <- varBag.Prior
            
            SendMessage = fun idx ->
                match idx with
                | 0 -> SendMessageHelper msgIdx (varIdx,varBag) 
                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
        }
        
        

    // Assuming that weights^T * var == 0 
    // This implementation uses cache, with the same interface as GaussianMultiSparseWeightedSumFactor
    
    // sparseWeights is (degree + 1) dimensional, where degree is the number of incoming edges.
    // The sparseWeights is sparse, stored as array<int * float>.  sparseWeights[0] is for the outgoing edge, including its id 0 and its weight.
    //  The other sparseWeights[i] have id = feature_index (from 0) and the corresponding weights

    // The varIdx and DistributionBag are (degree + 1) dimensional list of active variables, including the outgoing one as the first one [0].
    // takellh is (degree + 1) dimensional list of messages, including [0] for the outgoing node.    
    
    // This is NOT a generic implementation for multi weighted sum factors.  Let this factor be f with variables a1...an
    // Be careful with keeping the buffer up-to-date.
    // Whenever a node a_i receives a message from *other* factors, it should record that its potential has changed.
    // When computing message f -> a_j, the function below should look into all other adjacent variables a_i whose potential has changed since last msg sending,
    //  and update a_i's contribution to this factor's buffer actively
    // Now we know the schedule: when sending msg f -> weight var, only the discriminant variable has changed
    member this.GaussianFastSmallMultiSparseWeightedSumFactor (sparseWeights : array<int * float>) //(takellh : bool array) 
                                                (varIdx: array<int>) (distributionBagList : array<DistributionBag<Gaussian<'a>>>) 
                                                name scalingFactor =
        // the incoming degree
        let degree = sparseWeights.Length - 1;
        
        // only need to allocate degree + 1 number of messages
        let msgIndices = Array.init (degree + 1) (fun _ -> MessageBag.NewDistribution ())   // the only mapping
        let totalWeight = sparseWeights |> Array.map (fun (a, b) -> b)
        
        let acc_numer = ref 0.0<_> 
        let acc_denom = ref 0.0<_> 
        let discMsgIncluded = ref false 
        let discMsgSmall = ref false 
        let acc_logProdNorm = ref 0.0

        /// send message to idx and update its marginal (idx \in [0, degree])
        let UpdateHelper idx =
            let newMsg = 
                if idx = 0 then     // sending to discriminant variable
                    let (temp_numer, temp_denom) = [1..degree] |> Seq.fold (fun (acc_n, acc_d) edgeIdx ->                                      
                                      let tempGaussian = distributionBagList.[edgeIdx].[varIdx.[edgeIdx]] / MessageBag.[msgIndices.[edgeIdx]]
                                      let tempWeight = snd sparseWeights.[edgeIdx]
                                      (acc_n + tempWeight * tempGaussian.PrecisionMean / tempGaussian.Precision,
                                        acc_d + tempWeight * tempWeight / tempGaussian.Precision)                                  
                                                    ) (0.0<_>, 0.0<_>)
                    // outgoing message
                    acc_numer := temp_numer
                    acc_denom := temp_denom
                    Gaussian(!acc_numer / !acc_denom, 1.0 / !acc_denom)
                else
                    // First include the message that the discriminant variable received from difference factor
                    if not !discMsgIncluded then
                        discMsgIncluded := true
                        let gau = distributionBagList.[0].[varIdx.[0]] / MessageBag.[msgIndices.[0]]
                        if gau.Precision > zeroPrecisionTolerance * 1.0<_> then
                            acc_numer := !acc_numer - gau.PrecisionMean / gau.Precision
                            acc_denom := !acc_denom + 1.0 / gau.Precision
                        else
                            discMsgSmall := true    // no need to check every time.  Once is enough.

                    if !discMsgSmall then
                        Gaussian(0.0<_>, 0.0<_>)
                    else
                        //let featureIdx = 1 + fst sparseWeights.[idx]
                        let tempGaussian = distributionBagList.[idx].[varIdx.[idx]] / MessageBag.[msgIndices.[idx]]
                        let tempWeight = snd sparseWeights.[idx]
                        let denom = !acc_denom - tempWeight * tempWeight / tempGaussian.Precision
                        let numer = !acc_numer - tempWeight * tempGaussian.PrecisionMean / tempGaussian.Precision
                        Gaussian(- tempWeight * numer / denom * scalingFactor, 
                                   tempWeight * tempWeight / denom * scalingFactor)

            // receiverFeaIdx is the feature index of the receiving node, i.e, index in the array of distributionBagList and varIdx, so have to add 1
            //let receiverFeaIdx = if idx = 0 then 0 else 1 + fst sparseWeights.[idx]
            let oldMarginal = distributionBagList.[idx].[varIdx.[idx]]
            let newMarginal = oldMarginal / MessageBag.[msgIndices.[idx]] * newMsg
//            if printVarName then
//                newMarginal.Name <- oldMarginal.Name

            /// Update the message and marginal
            
            MessageBag.[msgIndices.[idx]] <- newMsg
            distributionBagList.[idx].[varIdx.[idx]] <- newMarginal

            if float newMarginal.Precision |> System.Double.IsNaN || float newMarginal.PrecisionMean |> System.Double.IsNaN then
                printfn "look GaussianFastSmallMultiSparseWeightedSumFactor"
            
            if debug && show_factor then
                printfn "In GaussianFastSmallMultiSparseWeightedSumFactor, Msg %s -> %s: (%A, %A), new marginal = (%A, %A)" name 
                                            (distributionBagList.[idx].[varIdx.[idx]].Name)
                                            newMsg.PrecisionMean newMsg.Precision newMarginal.PrecisionMean newMarginal.Precision

            /// buffer the contribution to the evidence
            if idx > 0 then
                acc_logProdNorm := !acc_logProdNorm + Gaussian.LogProductNormalisation (oldMarginal, newMsg)
            
            /// Return the difference between old and new marginals
            newMarginal - oldMarginal
            
        // returning an IFactor
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            
            Name = fun () -> name
            
            getType = fun () -> 6
            
            UpdateMessage = fun idx ->
                if idx >= 0 && idx <= degree then
                    UpdateHelper idx
                else
                    failwith (sprintf "Variable index out of range. Valid values are 0 to %d." degree)
            
            LogNormalisation = fun () -> 
                0.0
//                let result = [|1..degree|] |> Array.fold_left (fun acc idx ->
//                                acc + Gaussian.LogRatioNormalisation(distributionBagList.[idx].[varIdx.[idx]], MessageBag.[msgIndices.[idx]])
//                             ) !acc_logProdNorm
//                       
//                result

            NumberOfMessages = 1 // 
            
            ResetMarginals = fun () ->    
//                printfn "In Reset, %A, %A" distributionBagList.[0].[varIdx.[0]].PrecisionMean distributionBagList.[0].[varIdx.[0]].Precision
                distributionBagList.[0].[varIdx.[0]] <- distributionBagList.[0].Prior
//                [ 0..degree ] |> List.iter (fun idx ->                     
//                    distributionBagList.[idx].[varIdx.[idx]] <- distributionBagList.[idx].Prior
//                    )
            
            SendMessage = fun (idx:int) -> 
                0.0
//                if idx = 0 then
//                    let result = SendMessageHelper msgIndices.[0] (varIdx.[0], distributionBagList.[0])
////                    printfn "In SendMessage, %A, %A" distributionBagList.[0].[varIdx.[0]].PrecisionMean distributionBagList.[0].[varIdx.[0]].Precision
//                    result
//                elif idx <= degree then
//                    0.0 //
////                elif idx <= degree then
////                    0.0                    
//                else
//                    failwith (sprintf "Variable index out of range. Valid values are 0 to %d" degree)                
        }



    /// Connects a Gaussian noisy loss factor to variable and returns the update function
    /// if rho > 0.9 then it is noisy positive, if rho < 0.1 then it is noisy negative
    member this.GaussianNoisyFactor epsilon rho (varIdx, varBag:DistributionBag<Gaussian<'a>>) name =
    
        let msgIdx = MessageBag.NewDistribution ()
        let logEvidence = ref 0.0
                
        {
            ChangePrior = fun (arg_mu, arg_sigma2) -> ()
            getPrior = fun () -> (0.0<_>, 0.0<_>)
            Name = fun () -> name
            getType = fun () -> 7

            UpdateMessage = fun idx ->
                match idx with
                | 0 ->                 

                    let oldMarginal = varBag.[varIdx]
                    let oldMsg = MessageBag.[msgIdx]
                    let msgFromVar = oldMarginal / oldMsg

//                    printfn "msgFromVar = (%A, %A)" msgFromVar.Mean msgFromVar.StandardDeviation

                    let B = normcdf((-epsilon + msgFromVar.Mean) / msgFromVar.StandardDeviation)
                    let C = normcdf((-epsilon - msgFromVar.Mean) / msgFromVar.StandardDeviation)
                    let A = (1.0 - rho) * B + rho * C
                    let ratio1 = B * (1.0 - rho) / A
                    let ratio2 = C * rho / A
                    
                    let (newMarginal, newMsg) =
                        if ratio1 > 0.01 && ratio2 > 0.01 then
                            
//                            printfn "**** In lenient, ratio1 = %A, ratio2 = %A" ratio1 ratio2
                            
                            logEvidence := log (A)
                            let (_, _, mu_ge, var_ge) = GreaterThanStat epsilon msgFromVar
                            let (_, _, mu_le, var_le) = LessThanStat epsilon msgFromVar
                                                        
                            let mean_h = ratio1 * mu_ge + ratio2 * mu_le
                            let var_h = ratio1 * (mu_ge * mu_ge + var_ge) 
                                        + ratio2 * (mu_le * mu_le + var_le)
                                        - mean_h * mean_h

//                            printfn "A = %A, B = %A, C = %A, (%A, %A) mu_ge = %A\nvar_ge = %A, mu_le = %A, var_le = %A, mean_h = %A, var_h = %A" 
//                                            A B C ratio1 ratio2 mu_ge var_ge mu_le var_le mean_h var_h
//                            printfn "old marginal = (%A, %A)" oldMarginal.Mean oldMarginal.StandardDeviation
                            
                            let newMarginal = Gaussian.Create(mean_h, sqrt(var_h))
                            let newMsg = newMarginal / oldMarginal * oldMsg
//                            if newMsg.Precision < 0.0<_> then
//                                printfn "*** In lenient, %A, rho = %A, B = %A, C = %A, A = %A, ratio1 = %A, ratio2 = %A" 
//                                        newMsg.Precision rho B C A ratio1 ratio2
                            printfn "Oh, no"
                            
                            (newMarginal, newMsg)

                        elif ratio2 < 0.01 then 

//                            printfn "**** In ratio2 < 0.01, ratio1 = %A, ratio2 = %A" ratio1 ratio2

                            logEvidence := lognormcdf ((msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)
                            
                            let (newPrecisionMean, newPrecision, _, _) = GreaterThanStat epsilon msgFromVar                    
                            let newMsg = Gaussian(newPrecisionMean, newPrecision)
                            
                            if newMsg.Precision < 0.0<_> then
                                printfn "*** ratio2 < 0.01, %A" newMsg.Precision                               

                            (oldMarginal / oldMsg * newMsg, newMsg)
                            
                        else
//                            printfn "**** In ratio1 < 0.01, ratio1 = %A, ratio2 = %A" ratio1 ratio2

                            logEvidence := lognormcdf ((-msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)
                            
                            let (newPrecisionMean, newPrecision, _, _) = LessThanStat epsilon msgFromVar                    
                            let newMsg = Gaussian(newPrecisionMean, newPrecision)
                            
                            if newMsg.Precision < 0.0<_> then
                                printfn "*** In ratio1 < 0.01, %A" newMsg.Precision

                            (oldMarginal / oldMsg * newMsg, newMsg)
                    
//                    printfn "In GaussianNoisyFactor, newMsg = (%A, %A)" newMsg.PrecisionMean newMsg.Precision


                    /// Update the message and marginal
                    MessageBag.[msgIdx] <- newMsg
                    varBag.[varIdx] <- newMarginal
                    
//                    if Double.IsNaN (float newMsg.Precision) then
//                        printfn "ooooooo %A,  %A,  %A" newMarginal.Precision newMsg.Precision oldMarginal.Precision

                    /// Return the difference in the new marginal
                    newMarginal - oldMarginal                    
                | _ ->  failwith "Variable index out of range. Valid values are 0 only."


            LogNormalisation = fun () -> 
                !logEvidence
//                let marginal = varBag.[varIdx]
//                let msg = MessageBag.[msgIdx]
//                let msgFromVar = marginal / msg
//
//                let part1 = - Gaussian.LogProductNormalisation (msgFromVar, msg) 
//                let part2 = log((1.0 - rho) * normcdf ((msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation)
//                            + rho * normcdf ((-msgFromVar.Mean - epsilon)/msgFromVar.StandardDeviation))
//                            
////                printfn "%A + %A = %A" part1 part2 (part1 + part2)
//                
////                if Double.IsNaN part1 then
////                    printfn "Look: (%A, %A)" msgFromVar.Precision msg.Precision

//                part1 + part2

            NumberOfMessages = 1

            ResetMarginals = fun () ->
                varBag.[varIdx] <- varBag.Prior
            
            SendMessage = fun idx ->
                0.0
//                match idx with
//                | 0 -> SendMessageHelper msgIdx (varIdx,varBag) 
//                | _ ->  failwith "Variable index out of range. Valid values are 0 only."
        }            
